﻿namespace HeuristicLab.Algorithms.PushGP.Expressions {
  using System;
  using System.Collections.Generic;
  using System.Linq;

  public static class ExpressionTable {
    private static readonly Dictionary<string, Expression> statelessExpressionTable;

    private static readonly Dictionary<string, Func<Expression>> statefullExpressionFactory;

    public static int Count;

    static ExpressionTable() {
      statelessExpressionTable = GetStatelessExperssionTable(e => e.StringRepresentation);
      statefullExpressionFactory = GetStatefullExpressionFactory(e => e.StringRepresentation);

      Count = statelessExpressionTable.Count + statefullExpressionFactory.Count;
    }

    public static int StatelessCount
    {
      get
      {
        return statelessExpressionTable.Count;
      }
    }

    public static int StatefullCount
    {
      get
      {
        return statefullExpressionFactory.Count;
      }
    }

    private static Dictionary<string, Expression> GetStatelessExperssionTable(Func<Expression, string> keyFunc) {
      var dictionary = new Dictionary<string, Expression>();
      var expressionTypes = GetExpressionTypes(typeof(StatelessExpression));

      foreach (var type in expressionTypes) {
        var expression = Activator.CreateInstance(type) as Expression;
        var key = keyFunc(expression);

        dictionary.Add(key, expression);
      }

      return dictionary;
    }

    private static Dictionary<string, Func<Expression>> GetStatefullExpressionFactory(Func<Expression, string> keyFunc) {
      var dictionary = new Dictionary<string, Func<Expression>>();
      var statefullExpressionType = typeof(StatefullExpression<>);
      var expressionTypes = GetExpressionTypes(statefullExpressionType);

      foreach (var type in expressionTypes) {
        // Make a NewExpression that calls the ctor
        var newExp = System.Linq.Expressions.Expression.New(type);

        // Create a lambda with the New expression as body
        var creator = System.Linq.Expressions.Expression.Lambda<Func<Expression>>(newExp).Compile();
        var expression = creator();
        var key = keyFunc(expression);

        dictionary.Add(key, creator);
      }

      return dictionary;
    }

    private static bool IsSubclassOf(Type type, Type baseType) {
      if (type == null || baseType == null || type == baseType)
        return false;

      if (baseType.IsGenericType == false) {
        if (type.IsGenericType == false)
          return type.IsSubclassOf(baseType);
      } else {
        baseType = baseType.GetGenericTypeDefinition();
      }

      type = type.BaseType;
      var objectType = typeof(object);
      while (type != objectType && type != null) {
        var curentType = type.IsGenericType ? type.GetGenericTypeDefinition() : type;
        if (curentType == baseType)
          return true;

        type = type.BaseType;
      }

      return false;
    }

    private static IEnumerable<Type> GetExpressionTypes(Type baseType) {
      return from domainAssembly in AppDomain.CurrentDomain.GetAssemblies()
             from assemblyType in domainAssembly.GetTypes()
             where !assemblyType.IsAbstract &&
                   IsSubclassOf(assemblyType, baseType) &&
                   assemblyType.GetConstructor(Type.EmptyTypes) != null
             select assemblyType;
    }

    public static Expression GetExpression(int index) {
      // TODO: build lookup table for index to determine the key for the dictionary (get rid of ElementAt, is it slow?)
      return index < StatelessCount
               ? statelessExpressionTable.ElementAt(index).Value
               : statefullExpressionFactory.ElementAt(index - StatelessCount).Value();
    }

    public static Expression GetExpression(string symbol) {
      Expression expression;
      if (TryGetStatelessExpression(symbol, out expression) || TryGetStatefullExpression(symbol, out expression)) return expression;

      return null;
    }

    public static Expression GetStatelessExpression(string symbol) {
      Expression expression;
      if (statelessExpressionTable.TryGetValue(symbol, out expression)) return expression;
      throw new NotSupportedException("Expression not supported: " + symbol);
    }

    public static bool TryGetStatelessExpression(string name, out Expression expression) {
      return statelessExpressionTable.TryGetValue(name, out expression);
    }

    public static Expression GetStatefullExpression(string symbol) {
      Func<Expression> creator;
      if (statefullExpressionFactory.TryGetValue(symbol, out creator)) return creator();
      throw new NotSupportedException("Expression not supported: " + symbol);
    }

    public static bool TryGetStatefullExpression(string name, out Expression expression) {
      Func<Expression> creator;
      if (statefullExpressionFactory.TryGetValue(name, out creator)) {
        expression = creator();
        return true;
      }
      expression = default(Expression);
      return false;
    }
  }
}